<?php

namespace Drupal\uswds_storybook\Element;

use Drupal\Core\Render\Attribute\FormElement;
use Drupal\Core\Render\Element\Select;

/**
 * Provides a form element for a drop-down menu.
 *
 * Use this in place of radios or select (where multiple is false and there is
 * no optgroup hierarchy), particularly if there are more than 15 items. It
 * allows the user to begin typing, and any matching options will appear in a
 * selectable list beneath the combobox.
 *
 * Properties:
 * - #options: An key-value array of options to appear in the combobox, where
 *   the key is the "machine value," and the value is the human-readable value.
 *   There should not be any optgroup hierarchy here.
 * - #sort_options: (optional) If set to TRUE (default is TRUE), sort the
 *   options by their labels, after rendering and translation is complete.
 * - #sort_start: (optional) Option index to start sorting at, where 0 is the
 *   first option. If an empty option is being added automatically (see
 *   #empty_option and #empty_value properties), this defaults to 1 to keep the
 *   empty option at the top of the list. Otherwise, it defaults to 0.
 * - #empty_option: (optional) The label to show for the default option.
 *   By default, the label is automatically set to "- Select -" for a required
 *   field and "- None -" for an optional field.
 * - #empty_value: (optional) The value for the first default option, which is
 *   used to determine whether the user submitted a value or not. Defaults to
 *   the empty string.
 * - #default_value: Must be NULL or not set in case there is no value for the
 *   element yet, in which case a first default option is inserted by default.
 *   Whether this first option is a valid option depends on whether the field
 *   is #required or not.
 * - #required: (optional) Whether the user needs to select an option (TRUE)
 *   or not (FALSE). Defaults to FALSE.
 */
#[FormElement('combobox')]
class ComboBox extends Select {

  /**
   * {@inheritdoc}
   */
  public function getInfo() {
    $info = parent::getInfo();
    $info['#sort_options'] = TRUE;
    $info['#attached']['library'][] = 'uswds_storybook/combobox';
    return $info;
  }

  /**
   * {@inheritdoc}
   */
  public static function preRenderSelect($element) {
    parent::preRenderSelect($element);
    // Replace Drupal default class with USWDS-specific class.
    $index = array_search('form-select', $element['#attributes']['class']);
    if ($index === FALSE) {
      $index = count($element['#attributes']['class']);
    }
    $element['#attributes']['class'][$index] = 'usa-select';
    return $element;
  }

}
